# Copyright (C) 2024 - Scalizer (<https://www.scalizer.fr>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

from .record import OdooRecord

class OdooRecordSet(set):
    _odoo = None
    _model = None

    def __init__(self, seq=(), model=None):
        super().__init__(seq)
        if model:
            self._model = model
            self._odoo = self._model._odoo

    def __getattr__(self, name):
        if name == 'ids':
            return self.get_ids()
        else:
            return self.super().__getattr__(name)

    def save(self, batch_size=100, skip_line=0):
        values_list = [r.get_update_values() for r in self]
        self._model.load_batch(values_list, batch_size=batch_size, skip_line=skip_line)
        for r in self:
            r._updated_values = {}
        return True

    def get_ids(self):
        return [r.id for r in self]

    def unlink(self):
        ids = self.ids
        self._model.unlink(ids)
        for key in ids:
            self._model._cache.pop(key, None)
        self.clear()



# Copyright (C) 2024 - Scalizer (<https://www.scalizer.fr>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).

import base64
from datetime import datetime
import logging
import ssl
import os
import sys
import xmlrpc.client
from pprint import pformat

import requests
from bs4 import BeautifulSoup

from .model import OdooModel
from .record import OdooRecord
from .record_set import OdooRecordSet

METHODE_MAPPING = {
    15: [('get_object_reference', 'check_object_reference')]
}


class OdooConnection:
    _context = {'lang': 'fr_FR', 'noupdate': True}
    _models = {}
    query_count = 0
    method_count = {}

    def __init__(self, url, dbname, user, password, version=15.0, http_user=None, http_password=None, createdb=False,
                 debug_xmlrpc=False, legacy=False, logger=None):
        self.logger = logger or logging.getLogger("Odoo Connection".ljust(15))
        if debug_xmlrpc:
            self.logger.setLevel(logging.DEBUG)
        else:
            self.logger.setLevel(logging.INFO)
        self._url = url
        self._dbname = dbname
        self._user = user
        self._password = password
        self._http_user = http_user
        self._http_password = http_password
        self._version = version
        self._legacy = legacy
        self._debug_xmlrpc = debug_xmlrpc
        # noinspection PyProtectedMember,PyUnresolvedReferences
        self._insecure_context = ssl._create_unverified_context()
        self._compute_url()
        if createdb:
            self._create_db()
        self._prepare_connection()

    def __str__(self):
        return 'OdooConnection(%s)' % self._dbname

    def __repr__(self):
        return str(self)

    @property
    def context(self):
        return self._context

    def model(self, model_name):
        if model_name in self._models:
            return self._models[model_name]
        res = OdooModel(self, model_name)
        self._models[model_name] = res
        return res

    def _compute_url(self):
        if self._http_user or self._http_password:
            self._url = self._url.replace('https://', 'https://%s:%s@' % (self._http_user, self._http_password))

    def _get_xmlrpc_method(self, method):
        new_method = method
        for v in METHODE_MAPPING:
            if self._version >= v:
                for i in METHODE_MAPPING[v]:
                    if i[0] == method:
                        new_method = i[1]
        return new_method

    def init_logger(self, name='s6r-odoo', level='INFO'):
        root_logger = logging.getLogger()
        root_logger.setLevel(level)
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(logging.DEBUG)
        formatter = logging.Formatter(f'%(asctime)s - {name} - %(levelname)s - %(message)s')
        handler.setFormatter(formatter)
        root_logger.addHandler(handler)
        self.logger = root_logger

    def _create_db(self):
        post = {
            'master_pwd': "admin123",
            'name': self._dbname,
            'login': self._user,
            'password': self._password,
            'phone': '',
            'lang': 'fr_FR',
            'country_code': 'fr',
        }
        session = requests.Session()
        session.verify = False
        r = session.post(url=self._url + "/web/database/create", params=post)
        soup = BeautifulSoup(r.text, 'html.parser')
        alert = soup.find('div', attrs={"class": u"alert alert-danger"})
        if alert:
            self.logger.debug(self._url + "/web/database/create")
            self.logger.debug(post)
            self.logger.debug(alert.get_text())
            if "already exists" not in alert.text:
                raise Exception(alert.text)

    def _prepare_connection(self):
        self.logger.info("Prepare connection %s %s %s" % (self._url, self._dbname, self._user))
        self.common = xmlrpc.client.ServerProxy('{}/xmlrpc/2/common'.format(self._url), allow_none=True,
                                                context=self._insecure_context)
        self.object = xmlrpc.client.ServerProxy('{}/xmlrpc/2/object'.format(self._url), allow_none=True,
                                                context=self._insecure_context)
        self.logger.info("==============")
        try:
            self.uid = self.common.authenticate(self._dbname, self._user, self._password, {})
        except xmlrpc.client.Fault as err:
            if 'FATAL:' in err.faultString:
                msg = err.faultString[err.faultString.find('FATAL:') + 6:].strip()
                self.logger.error(msg)
                raise ConnectionError(msg)
            raise err
        except xmlrpc.client.ProtocolError as err:
            msg = f'{err.url} {err.errmsg} ({err.errcode})'
            self.logger.error(msg)
            raise ConnectionError(msg)
        except Exception as err:
            self.logger.error(err)
            raise

        self.logger.debug('Connection uid : %s' % self.uid)
        if not self.uid:
            msg = f'Connection Error to {self._url} {self._dbname}. Check "{self._user}" username and password.'
            self.logger.error(msg)
            raise ConnectionError(msg)

    def query_counter_update(self, model, method):
        self.query_count += 1
        if method not in self.method_count:
            self.method_count[method] = {model: 1}
        else:
            if model not in self.method_count[method]:
                self.method_count[method][model] = 1
            else:
                self.method_count[method][model] += 1

    def execute_odoo(self, *args, no_raise=False):
        model = args[0]
        method = args[1]
        self.query_counter_update(model, method)

        self.logger.debug("*" * 50)
        self.logger.debug("Execute odoo :")
        self.logger.debug("\t Model : %s" % model)
        self.logger.debug("\t Method : %s" % method)
        self.logger.debug("\t " + "%s " * (len(args) - 2) % args[2:])
        self.logger.debug("*" * 50)
        try:
            res = self.object.execute_kw(self._dbname, self.uid, self._password, *args)
            return res
        except Exception as e:
            if no_raise:
                pass
            else:
                self.logger.error(pformat(args))
                self.logger.error(e)
                raise e

    def values_to_record(self, model_name, values, update_cache=True):
        if isinstance(values, int):
            values = {'id': values}
        record = OdooRecord(self, self.model(model_name), values)
        if update_cache:
            self._models[model_name]._update_cache(record.id, values)
        return record

    def values_list_to_records(self, model_name, val_list, update_cache=True):
        if self._legacy:
            return val_list
        records = [self.values_to_record(model_name, values, update_cache) for values in val_list]
        if not records:
            return []
        if len(records) == 1:
            return records
        else:
            return OdooRecordSet(records, model=self.model(model_name))

    def get_ref(self, external_id):
        res = self.get_object_reference(external_id)[1]
        self.logger.debug('Get ref %s > %s' % (external_id, res))
        return res

    def get_local_file(self, path, encode=False):
        if encode:
            with open(path, "rb") as f:
                res = f.read()
            res = base64.b64encode(res).decode("utf-8", "ignore")
        else:
            with open(path, "r") as f:
                res = f.read()
        return res

    def get_country(self, code):
        return self.execute_odoo('res.country', 'search', [[('code', '=', code)], 0, 1, "id", False],
                                 {'context': self._context})[0]

    def get_menu(self, website_id, url):
        return self.execute_odoo('website.menu', 'search',
                                 [[('website_id', '=', website_id), ('url', '=', url)], 0, 1, "id", False],
                                 {'context': self._context})[0]

    def get_search_id(self, model, domain):
        return self.execute_odoo(model, 'search', [domain, 0, 1, "id", False], {'context': self._context})[0]

    def search_ir_model_data(self, domain):
        return self.model('ir.model.data').search(domain,
                                                  fields=['module', 'name', 'model', 'res_id'],
                                                  order='id')

    def get_object_reference_legacy(self, xml_id, no_raise=False):
        object_reference = self._get_xmlrpc_method('get_object_reference')
        return self.execute_odoo('ir.model.data', object_reference, xml_id.split('.'), no_raise=no_raise)


    def get_object_reference(self, xml_id, no_raise=False):
        if self._legacy:
            return self.get_object_reference_legacy(xml_id, no_raise=no_raise)

        ref = self._get_object_reference_cache(xml_id)
        if ref:
            return ref
        module, name = xml_id.split('.')
        domain = [('module', '=', module), ('name', '=', name)]
        res = self.search_ir_model_data(domain)
        if res:
            ir_model_data_id = res[0]
            return [ir_model_data_id.model, ir_model_data_id.res_id]

    def get_id_from_xml_id(self, xml_id, no_raise=False):
        if '.' not in xml_id:
            xml_id = "external_config." + xml_id
        try:

            res = self.get_object_reference(xml_id, no_raise=no_raise)
            return res[1] if res else False
        except xmlrpc.client.Fault as fault:
            if no_raise:
                pass
            raise ValueError(fault.faultString.strip().split('\n')[-1])
        except Exception as err:
            if no_raise:
                pass
            else:
                raise err

    def ref(self, ixml_id, no_raise=False, fields=None, no_cache=False):
        object_reference = self.get_object_reference(ixml_id, no_raise=no_raise)
        if len(object_reference) == 2:
            model, res_id = object_reference
            return self.model(model).read(res_id, fields=fields, no_cache=no_cache)

    def get_xml_id_from_id(self, model, res_id):
        cache_xmlid = self._get_xmlid_cache(model, res_id)
        if cache_xmlid:
            return cache_xmlid
        try:
            domain = [('model', '=', model), ('res_id', '=', res_id)]
            res = self.search_ir_model_data(domain)
            if res:
                datas = res[0]
                return "%s.%s" % (datas['module'], datas['name'])
            else:
                raise ValueError('xml_id not found.')
        except Exception as err:
            raise err

    def write(self, model, ids, values, context=None):
        return self.execute_odoo(model, 'write', [ids, values],
                                 {'context': context or self._context})

    def set_active(self, is_active, model, domain, search_value_xml_id):
        if search_value_xml_id:
            object_id = self.get_id_from_xml_id(search_value_xml_id)
            domain = [(domain[0][0], domain[0][1], object_id)]
        object_ids = self.search_ids(model, domain, context=self._context)
        self.write(model, object_ids, {'active': is_active})

    def read_search(self, model, domain, context=None):
        res = self.execute_odoo(model, 'search_read', [domain],
                                {'context': context or self._context})
        return res

    def search_count(self, model, domain, context=None):
        res = self.execute_odoo(model, 'search_count', [domain],
                                {'context': context or self._context})
        return res

    def _read(self, model, ids, fields, context=None):
        param_ids = [ids] if isinstance(ids, int) else ids
        return self.execute_odoo(model, 'read', [param_ids, fields],
                                 {'context': context or self._context})

    def read(self, model, ids, fields, context=None):
        res =  self._read(model, ids, fields, context)
        return self.values_list_to_records(model, res)

    def read_group(self, model, domain, fields, groupby, offset=0, limit=None, orderby=False, lazy=True, context=None):
        res = self.execute_odoo(model, 'read_group', [domain, fields, groupby, offset, limit, orderby, lazy],
                                {'context': context or self._context})
        return res

    def search(self, model, domain=[], fields=[], order="", offset=0, limit=0, context=None):
        params = [domain, fields, offset, limit, order]
        res = self.execute_odoo(model, 'search_read', params, {'context': context or self._context})
        return self.values_list_to_records(model, res)

    def search_ids(self, model, domain=[], order="", offset=0, limit=0, context=None):
        params = [domain, offset, limit, order]
        res = self.execute_odoo(model, 'search', params, {'context': context or self._context})
        return res

    def get_record(self, model, rec_id, context=None):
        params = [[('id', '=', rec_id)]]
        res = self.execute_odoo(model, 'search_read', params,
                                {'context': context or self._context})
        if res:
            return res[0]

    def default_get(self, model, field):
        res = self.execute_odoo(model, 'default_get', [field])
        return res

    def load(self, model, load_keys, load_data, context):
        res = self.execute_odoo(model, 'load', [load_keys, load_data], {'context': context or self._context})
        for message in res['messages']:
            self.logger.error("%s : %s" % (message['record'], message['message']))
        return res

    def load_batch(self, model, datas, batch_size=100, skip_line=0, context=None, ignore_fields=[]):
        context = self.context | context if context else self.context
        if not datas:
            return
        cc_max = len(datas)
        start = datetime.now()

        load_keys = list(datas[0].keys())
        for field in ignore_fields:
            try:
                load_keys.remove(field)
            except ValueError:
                self.logger.warning(f"\"{field}\" field name not found in data keys")
        load_datas = [[]]
        for cc, data in enumerate(datas):
            if len(load_datas[-1]) >= batch_size:
                load_datas.append([])
            load_datas[-1].append([data[i] for i in load_keys])

        cc = 0
        for load_data in load_datas:
            start_batch = datetime.now()
            self.logger.info("\t\t* %s : %s-%s/%s" % (model, skip_line + cc, skip_line + cc + len(load_data), skip_line + cc_max))
            cc += len(load_data)
            res = self.load(model, load_keys, load_data, context=context)
            for message in res['messages']:
                if message.get('type') in ['warning', 'error']:
                    if message.get('record'):
                        self.logger.error("record : %s" % (message['record']))
                    if message.get('message'):
                        self.logger.error("message : %s" % (message['message']))
                    raise Exception(message['message'])
                else:
                    self.logger.info(message)
            stop_batch = datetime.now()
            self.logger.info("\t\t\tBatch time %s ( %sms per object)" % (
                stop_batch - start_batch, ((stop_batch - start_batch) / len(load_data)).microseconds / 1000))
        stop = datetime.now()
        self.logger.info("\t\t\tTotal time %s" % (stop - start))

    def create(self, model, values, context=None):
        res = self.execute_odoo(model, 'create', [values],  {'context': context or self._context})
        return self.values_list_to_records(model, res)

    def unlink(self, model, values, context=None):
        return self.execute_odoo(model, 'unlink', [values],  {'context': context or self._context})

    def unlink_domain(self, model, domain, context=None):
        values = self.search_ids(model, domain)
        return self.unlink(model, values, context)

    def create_attachment(self, name, datas, res_model, res_id=False, context=None):
        values = {
                    'name': name,
                    'datas': datas,
                    'res_model': res_model}
        if res_id:
            values['res_id'] = res_id
        return self.create('ir.attachment', values,  context)

    def create_attachment_from_local_file(self, file_path, res_model, res_id=False,
                                          name=False, encode=False, context=None):
        datas = self.get_local_file(file_path, encode)
        file_name = name or os.path.basename(file_path)
        return self.create_attachment(file_name, datas, res_model, res_id, context)

    def get_ir_model_data(self, model):
        return self.search('ir.model.data', [('model', '=', model)],
                    fields=['module', 'name', 'model', 'res_id'])

    def get_id_ref_dict(self, model):
        """
        Returns a dict with xmlid as key and id as value
        :param model: Model name
        :return: {'base.module_account': 894, ...}
        """
        model_datas = self.get_ir_model_data(model)
        return dict([(data.res_id, '%s.%s' % (data.module, data.name)) for data in model_datas])

    def get_xmlid_dict(self, model):
        """
        Returns a dict with id as key and xmlid as value
        :param model: Model name
        :return: {894: 'base.module_account', ...}
        """
        model_datas = self.get_ir_model_data(model)
        return dict([('%s.%s' % (data.module, data.name), data.res_id) for data in model_datas])

    def get_fields(self, model, fields=None, attributes=None):
        if model in self._models and self._models[model]._fields and not fields and not attributes:
            return self._models[model]._fields
        params = [fields or []]
        if attributes:
            params.append(attributes)

        return self.execute_odoo(model, 'fields_get', params)

    def _get_xmlid_cache(self, model, res_id):
        if not 'ir.model.data' in self._models:
            return
        cache = self._get_model_cache('ir.model.data')
        ir_model_datas = list(filter(lambda x: x['model'] == model and x['res_id'] == res_id, cache))
        if ir_model_datas:
            return '{0}.{1}'.format(ir_model_datas[0]['module'], ir_model_datas[0]['name'])

    def _get_object_reference_cache(self, xml_id):
        if not 'ir.model.data' in self._models:
            return
        module, name = xml_id.split('.')
        cache = self._get_model_cache('ir.model.data')
        ir_model_datas = list(filter(lambda x: x['module'] == module and x['name'] == name, cache))
        if ir_model_datas:
            return [ir_model_datas[0]['model'], ir_model_datas[0]['res_id']]

    def _get_model_cache(self, model):
        cache = self._models[model]._cache
        return [cache[res_id] for res_id in cache.keys()]

    def print_query_count(self):
        self.logger.info('Query count : %s', self.query_count)
        for method in self.method_count:
            self.logger.info('Method Count %s : %s', method, self.method_count[method])



# Copyright (C) 2024 - Scalizer (<https://www.scalizer.fr>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).


class OdooRecord(object):
    _odoo = None
    _model = None
    _field = ''
    _parent_model = None

    def __init__(self, odoo, model, values: dict, field='', parent_model=None):
        self._values = {}
        self._updated_values = {}
        self._initialized_fields = []
        self._odoo = odoo
        self.id = False
        if model:
            self._model = model
            self._odoo = self._model._odoo
        if field:
            self._field = field
        if parent_model:
            self._parent_model = parent_model

        self.set_values(values, update_cache=False)

    def __str__(self):
        if self._model:
            if hasattr(self, 'id'):
                return "%s(%s)" % (self._model, self.id)
            if hasattr(self, 'name'):
                return "%s(%s)" % (self._model, self.name)
        elif self._field:
            if hasattr(self, 'id'):
                return "%s(%s)" % (self._field, self.id)
            if hasattr(self, 'name'):
                return "%s(%s)" % (self._field, self.name)
        return str(self._values)

    def __repr__(self):
        return str(self)

    def __bool__(self):
        if hasattr(self, 'id'):
            return bool(self.id)

    def __getitem__(self, key):
        if isinstance(key, str):
            return getattr(self, key)

    def __getattr__(self, name):
        if name.startswith('_'):
            return self.super().__getattr__(name)
        if not self._model:
            return self.super().__getattr__(name)
        if name not in self._values:
            if not self._model._fields_loaded:
                self._model.load_fields_description()
            if  name in self._model._fields:
                self.read([name])
                return getattr(self, name)
            raise AttributeError("Attribute '%s' not found in model '%s'" % (name, self._model))

    def __setattr__(self, name, value):
        if name.startswith('_'):
            return super().__setattr__(name, value)
        if name in self._values and name in self._initialized_fields and value != self._values[name]:
            self._updated_values[name] = value
            return super().__setattr__(name, value)
        else:
            res = super().__setattr__(name, value)
            self._initialized_fields.append(name)
            return res

    def _update_cache(self):
        if self._model:
            self._model._update_cache(self._values['id'], self._values)

    def set_values(self, values, update_cache=True):
        self._values.update(values)
        if self._model and update_cache:
            self._update_cache()
        for key in values:
            value = values[key]
            if isinstance(value, list) and len(value) == 2:
                setattr(self, key, OdooRecord(self._odoo, None, {'id': value[0], 'name': value[1]}, key, self._model))
            else:
                setattr(self, key, value)

    def read(self, fields=None, no_cache=False):
        if not self._model._fields_loaded:
            self._model.load_fields_description()
        if self.id in self._model._cache and not no_cache:
            res = self._model._cache[self.id]
            # check if all fields are in res dict
            if any(field not in res for field in fields):
                res.update(self._read(fields))

            self.set_values(res)
        else:
            if not fields:
                fields = self._model.get_fields_list()
            res = self._read(fields)
            if res:
                self.set_values(res)

    def _read(self, fields):
        res = self._model._read(self.id, fields)
        if res:
            return res[0]

    def save(self):
        if self.id:
            self._model.write(self.id, self._updated_values)
            self._updated_values = {}
        else:
            self.id = self._odoo._create(self._model.model_name, self._values)
            self._initialized_fields = list(self._values.keys())

    def write(self, values):
        self._model.write(self.id, values)
        self._values.update(values)
        self.__dict__.update(values)

    def refresh(self):
        self.read(self._initialized_fields, no_cache=True)

    def get_update_values(self):
        values = self._updated_values
        if self.id:
            values['.id'] = self.id
        else:
            values['id'] = self.id
        return values



# Copyright (C) 2024 - Scalizer (<https://www.scalizer.fr>).
# License LGPL-3.0 or later (https://www.gnu.org/licenses/lgpl.html).


class OdooModel(object):

    _fields_loaded = False
    _fields = {}
    _cache = {}
    _xmlid_cache = {}

    def __init__(self, odoo, model_name):
        self.model_name = model_name
        self._odoo = odoo
        self._cache = {}
        self._fields = {}
        self._xmlid_cache = {}

    def __str__(self):
        return self.model_name

    def __repr__(self):
        return str(self)

    def _update_cache(self,record_id, values):
        self._cache[record_id] = values

    def values_to_record(self, values, update_cache=True):
        if update_cache:
            self._update_cache(values['id'], values)
        return self._odoo.values_to_record(self.model_name, values, update_cache=False)

    def _get_cache(self, record_id):
        if record_id in self._cache:
            return self.values_to_record(self._cache.get(record_id), update_cache=False)

    def execute(self, model_method,  *args, no_raise=False):
        return self._odoo.execute_odoo(self.model_name, model_method, args, no_raise=no_raise)

    def search_get_id(self, domain):
        return self._odoo.get_search_id(self.model_name, domain)

    def get_xml_id_from_id(self, xml_id):
        return self._odoo.get_xml_id_from_id(self.model_name, xml_id)

    def read(self, ids, fields=None, context=None, no_cache=False):
        if not ids:
            return []
        if not fields:
            fields = self.get_fields_list()
        if isinstance(ids, int):
            if not no_cache:
                cache_record = self._get_cache(ids)
                if cache_record:
                    return cache_record

            record = self._odoo.read(self.model_name, [ids], fields, context)[0]
            if record:
                return record

        return self._odoo.read(self.model_name, ids, fields, context)


    def _read(self, ids, fields=None, context=None, no_cache=False):
        return self._odoo._read(self.model_name, ids, fields, context)

    def read_search(self, domain, context=None):
        return self._odoo.read_search(self.model_name, domain, context)

    def search_count(self, domain, context=None):
        return self._odoo.search_count(self.model_name, domain, context)

    def read_group(self, domain, fields, groupby, offset=0, limit=None, orderby=False, lazy=True, context=None):
        return self._odoo.read_group(self.model_name, domain, fields, groupby, offset, limit, orderby, lazy, context)

    def search(self, domain=[], fields=[], order="", offset=0, limit=0, context=None):
        return self._odoo.search(self.model_name, domain, fields, order, offset, limit, context)

    def search_ids(self, domain=[], fields=[], order="", offset=0, limit=0, context=None):
        return self._odoo.search_ids(self.model_name, domain, fields, order, offset, limit, context)

    def get_record(self, rec_id, context=None):
        return self._odoo.get_record(self.model_name, rec_id, context)

    def default_get(self, field):
        return self._odoo.default_get(self.model_name, field)

    def load(self, load_keys, load_data, context=None):
        return self._odoo.load(self.model_name, load_keys, load_data, context)

    def load_batch(self, data, batch_size=100, skip_line=0, context=None, ignore_fields=[]):
        return self._odoo.load_batch(self.model_name, data, batch_size, skip_line, context, ignore_fields)

    def write(self, ids, values, context=None):
        return self._odoo.write(self.model_name, ids, values, context)

    def create(self, values, context=None):
        return self._odoo.create(self.model_name, values, context)

    def unlink(self, values, context=None):
        return self._odoo.unlink(self.model_name, values, context)

    def unlink_domain(self, domain, context=None):
        return self._odoo.unlink_domain(self.model_name, domain, context)

    def create_attachment(self, name, datas, res_id=False, context=None):
        return self._odoo.create_attachment(name, datas, self.model_name, res_id, context)

    def create_attachment_from_local_file(self, file_path, res_id=False, name=False, encode=False, context=None):
        return self._odoo.create_attachment_from_local_file(file_path, self.model_name, res_id, name, encode, context)

    def get_id_ref_dict(self):
        return self._odoo.get_id_ref_dict(self.model_name)

    def get_xmlid_dict(self):
        return self._odoo.get_xmlid_dict(self.model_name)

    def get_fields(self, fields=None, attributes=None):
        return self._odoo.get_fields(self.model_name, fields, attributes)

    def load_fields_description(self):
        if not self._fields_loaded:
            self._fields = self.get_fields()
            self._fields_loaded = True

    def load_field_description(self, field):
        if field in self._fields:
            return self._fields[field]
        field_desc = self.get_fields([field])
        self._fields.update(field_desc)
        return field_desc[field]

    def get_fields_list(self):
        self.load_fields_description()
        return list(self._fields.keys())

    def get_ir_model_data(self):
        ir_model_datas = self._odoo.get_ir_model_data(self.model_name)
        self._xmlid_cache = ir_model_datas
        return ir_model_datas

    def _find_id_in_xmlid_cache(self, xml_id):
        module, name = xml_id.split('.')
        ir_model_datas = list(filter(lambda x: x.module == module and x.name == name, self._xmlid_cache))
        if ir_model_datas:
            return ir_model_datas[0].res_id

    def _find_ref_in_xmlid_cache(self, res_id):
        ir_model_datas = list(filter(lambda x: x.res_id == res_id, self._xmlid_cache))
        if ir_model_datas:
            return '{0}.{1}'.format(ir_model_datas[0].module, ir_model_datas[0].name)



import logging
import time
import random
import string
from s6r_odoo import OdooConnection


def get_random_string(length):
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for _ in range(length))

logging.basicConfig()
logger = logging.getLogger("test")
odoo = OdooConnection(url='http://odoo_digitalportage2.localhost',
                          dbname='digitalportage_prod',
                          user='admin',
                          password='admin', debug_xmlrpc=False, logger=logger)


xmlid_dict = odoo.model('ir.module.module').get_xmlid_dict()
logger.info('ir.module.module XMLIDs : %s', xmlid_dict)
id_ref_dict = odoo.model('ir.module.module').get_id_ref_dict()
logger.info('ir.module.module ID refs : %s', id_ref_dict)
ir_model_data = odoo.model('ir.module.module').get_ir_model_data()
logger.info('ir.module.module ir_model_data : %s', ir_model_data)


object_reference = odoo.get_object_reference('base.module_account')
logger.info("base.module_account get_object_reference --> ['ir.module.module', id] : %s", object_reference)
res_id = odoo.get_ref('base.module_account')
logger.info("base.module_account get_ref --> id : %s", res_id)
module_account = odoo.ref('base.module_account')
msg = 'base.module_account ref --> OdooRecord :\nModule: %s (%s)\nDescription: %s'
logger.info(msg, module_account.display_name,
            module_account.name,
            module_account.description)

odoo.get_xml_id_from_id('ir.module.module', module_account.id)

odoo.print_query_count()

partner_fields = odoo.execute_odoo('res.partner', 'fields_get', [['name'], None])
partner_fields2 = odoo.model('res.partner').execute('fields_get', ['name'])

country_fields = odoo.model('res.country').get_fields([])
country_required_fields = ', '.join(country_fields[key]['name'] for key in country_fields if country_fields[key]['required'])
logger.info('Country required fields : %s', country_required_fields)


#Script optimised to access license and category_id field
start_time = time.time()
module_ids = odoo.model('ir.module.module').search([('author', 'ilike', 'Odoo')],
                                                   fields=['display_name', 'category_id', 'license'],
                                                   order='create_date desc, name', limit=100)
optimized_query_time = time.time() - start_time
start_time = time.time()
logger.info('Module categories : %s', ', '.join(set([module.category_id.name for module in module_ids])))
for module_id in module_ids:
    module_license = module_id.license if module_id['license'] else 'Unknown'
    logger.info('* Module %s : %s', module_id.display_name, module_id.license)
    continue
optimized_time = time.time() - start_time
logger.info("Optimised Script : %s seconds (first query : %s seconds)", optimized_time, optimized_query_time)


# Script NOT optimised to access to licence and category_id field
# start_time = time.time()
# module_ids = odoo.model('ir.module.module').search([('author', 'ilike', 'Odoo')],
#                                                    fields=['display_name'],
#                                                    order='create_date desc, name', limit=10)
# not_optimized_query_time = time.time() - start_time
# start_time = time.time()
# logger.info('Module categories : %s', ', '.join(set([module.category_id.name for module in module_ids])))
# for module_id in filter(lambda x: x.category_id.name != 'Scalizer', module_ids): # module_ids:
#     module_license = module_id.license if module_id['license'] else 'Unknown'
#     logger.info('* Module %s : %s', module_id.display_name, module_id.license)
#     continue
# not_optimized_time = time.time() - start_time
# logger.info("NOT Optimised Script : %s seconds (first query : %s seconds)", not_optimized_time, not_optimized_query_time)

# XMLIDs Tests
logger.info('Module XMLIDs : %s', odoo.get_xmlid_dict('ir.module.module'))

object_id = odoo.get_ref('base.module_account')
logger.info('base.module_account get_ref --> id : %s', object_id)
module_account_lazy = odoo.model('ir.module.module').read(object_id, ['name', 'description'])

logger.info('Module (LAZY): %s\nCategory: %s\nDescription: %s\n',
            module_account_lazy['name'],
            module_account_lazy['category_id'].name,
            module_account_lazy['description'])

# test cache
origin_email = 'test@example.com'
email1 = 'john@example.com'
partner_id = odoo.ref('base.main_partner')
partner_id.email = email1
partner_id.save()
if partner_id.email != email1:
    raise Exception('Email %s not updated' % email1)
email2 = 'sophie@example.com'
partner_id.write({'email': email2})
if partner_id.email != email2:
    raise Exception('Email %s not updated' % email2)

partner_id2 = odoo.model('res.partner').read(partner_id.id)
if email2 in partner_id2.email_formatted:
    partner_id.write({'email': origin_email})
    raise Exception('Cache not used with read')

partner_id2 = odoo.ref('base.main_partner')
if email2 in partner_id2.email_formatted:
    partner_id.write({'email': origin_email})
    raise Exception('Cache not used with ref')

partner_id3 = odoo.ref('base.main_partner', no_cache=True)
if email2 not in partner_id3.email_formatted:
    partner_id.write({'email': origin_email})
    raise Exception('Cache used with ref')
partner_id3.write({'email': email1})
partner_id3.refresh()
if email1 not in partner_id3.email_formatted:
    partner_id.write({'email': origin_email})
    raise Exception('Refresh failed')

partner_id.write({'email': origin_email})

partner_ids = odoo.model('res.partner').search([], fields=['name', 'website'])
for partner_id in partner_ids:
    partner_id.website = 'https://www.%s.com' % get_random_string(8)
partner_ids.save()

partner_ids = odoo.model('res.partner').search([('name', '=', 'TO_REMOVE')], fields=['name'])
if partner_ids:
    partner_ids.unlink()

new_partner_values_list = [{'name': "TO_REMOVE", 'website': get_random_string(8)} for i in range(50)]
new_partner_ids = odoo.model('res.partner').create(new_partner_values_list)
new_partner_ids.unlink()

odoo.print_query_count()